-------------------------------------------------------------------------------
-- Lorewalkers by Lockslap, US-Bleeding Hollow
-- <Borderline Amazing>, http://ba-guild.com
-------------------------------------------------------------------------------
local Lorewalkers = LibStub("AceAddon-3.0"):NewAddon("Lorewalkers", "AceEvent-3.0", "AceConsole-3.0", "AceTimer-3.0")
if not TomTom then return end

-- libaries
local L = LibStub("AceLocale-3.0"):GetLocale("Lorewalkers", false)
local Astrolabe = DongleStub("Astrolabe-1.0")
local AceConfig = LibStub("AceConfig-3.0")
local AceConfigDialog = LibStub("AceConfigDialog-3.0")
local AceConfigRegistry = LibStub("AceConfigRegistry-3.0")

-- waypoints frame
local LorewalkersFrame = CreateFrame("frame", "LorewalkersFrame", UIParent)
local LorewalkersFrameBG = LorewalkersFrame:CreateTexture("LorewalkersFrameBG", "BACKGROUND")
LorewalkersFrame:EnableMouse(true)
LorewalkersFrame:SetResizable(false)
LorewalkersFrame:SetMovable(true)
LorewalkersFrame:SetUserPlaced(true)
LorewalkersFrame:SetPoint("CENTER")
LorewalkersFrameBG:SetAllPoints()
LorewalkersFrame:SetScript("OnDragStop", function(self) self:StopMovingOrSizing(); self:SetUserPlaced(true); end)
LorewalkersFrame:SetScript("OnMouseUp", function(self) self:StopMovingOrSizing(); self:SetUserPlaced(true); end)

-- setup some globals
Lorewalkers.name = GetAddOnMetadata("Lorewalkers", "Title")
Lorewalkers.version = GetAddOnMetadata("Lorewalkers", "Version")

local waypoint
local zones, db, waypoints = {}, {}, {}

local enabledString = ("|cff00ff00%s|r"):format(L["Enabled"])
local disabledString = ("|cffff0000%s|r"):format(L["Disabled"])
local repStandings 		= {
	[1]	= L["Hated"],
	[2]	= L["Hostile"],
	[3]	= L["Unfriendly"],
	[4]	= L["Neutral"],
	[5]	= L["Friendly"],
	[6]	= L["Honored"],
	[7]	= L["Revered"],
	[8]	= L["Exalted"],
}

-- options page
local fields = {"Title", "Version", "Author", "X-Category", "X-License", "X-Email", "Email", "eMail", "X-Website", "X-Credits", "X-Localizations", "X-Donate", "X-Paypal"}
local fieldName = "|cff00ff96%s:|r"

-- default settings
local defaults 		= {
	char = {
		enabled = true,
		notices = false, -- ditto for notices
		info = true,
		infoType = "criteria",
		arrow = true,
		minimap = true,
		worldmap = true,
		
		-- waypoint frame
		frame = {
			enabled = true,
			-- frame settings
			locked = false,
			scale = 1.0,
			alpha = 1.0,
		},
		
		-- ldb settings
		ldb = {
			display = "incomplete",
			combat = true,
			location = true,
			remaining = true,
			totalRemaining = true,
		},
		
		-- not used in config frame
		firstLoad = false,	-- for building our waypoint db
		waypoints = {},
	}
}

-- options table
local options = {
	type = "group",
	args = {
		general = {
			type = "group",
			name = Lorewalkers.name,
			get = function(info) return db[info[#info]] end,
			set = function(info, value) 
				db[info[#info]] = value
				Lorewalkers:UpdateFrame()
				if waypoint then TomTom:RemoveWaypoint(waypoint) end
				if db.enabled then
					Lorewalkers:AddClosestWaypoint()
				end
			end,
			args = {
				header = {
					type = "header",
					order = 1,
					name = (L["Version: |cff00ff96%s|r"]):format(Lorewalkers.version),
					width = "full",
				},
				description = {
					type = "description",
					order = 2,
					name = L["A simple addon to aid you with obtaining exalted with The Lorewalkers."],
					width = "full",
				},
				enabled = {
					type = "toggle",
					order = 3,
					name = L["Enable Lorewalkers"],
					desc = L["Toggle the waypoints."],
					width = "full",
					get = function() return db.enabled end,
					set = function(_, value)
						db.enabled = value
						Lorewalkers:UpdateFrame()
						if waypoint then TomTom:RemoveWaypoint(waypoint) end
						if not db.enabled then
							if LorewalkersFrame:IsShown() then LorewalkersFrame:Hide() end
						else
							Lorewalkers:AddClosestWaypoint()
						end
					end,
				},
				notices = {
					type = "toggle",
					order = 4,
					name = L["Display Notices"],
					desc = L["Notify you of various actions being taken by the addon.  For example, loading a new waypoint."],
				},
				info = {
					type = "toggle",
					order = 5,
					name = L["Print Detailed Information"],
					desc = L["Print detailed information about the criteria or zone waypoints when you complete an achievement's criteria."],
				},
				infoType = {
					type = "select",
					order = 6,
					name = L["Information Type"],
					desc = L["What type of information do you want to print when you complete an achievement's criteria?\n\n|cff00ff96Criteria:|r\nPrint the remaining number of criteria needed to complete the achievement.\n\n|cff00ff96Zone:|r\nPrint the number of waypoints remaining in the zone."],
					disabled = function() return not db.info end,
					values = {
						["criteria"] = L["Criteria"],
						["zone"] = L["Zone"],
					},
				},
				dummy = {
					type = "description",
					order = 99,
					name = "",
					width = "full",
				},
				addTo = {
					type = "group",
					order = 100,
					guiInline = true,
					name = L["Add To"],
					args = {
						arrow = {
							type = "toggle",
							order = 1,
							name = L["Crazy Arrow"],
							desc = L["Add waypoints to TomTom's crazy arrow."],
						},
						minimap = {
							type = "toggle",
							order = 2,
							name = L["Minimap"],
							desc = L["Add waypoints to the minimap."],
						},
						worldmap = {
							type = "toggle",
							order = 3,
							name = L["World Map"],
							desc = L["Add waypoints to the world map."],
						},
					},
				},
			},
		},
		ldb = {
			type = "group",
			order = 99,
			name = L["LDB Plugin"],
			get = function(info) return db.ldb[info[#info]] end,
			set = function(info, value) db.ldb[info[#info]] = value; Lorewalkers.LDB.text = LorewalkersGetRemainingCriteria() end,
			args = {
				display = {
					type = "select",
					order = 1,
					name = L["Display Type"],
					desc = L["What to show on the LDB text."],
					values = {
						["complete"] = L["Criteria Completed (Zone)"],
						["totalIncomplete"] = L["Criteria Completed (Total)"],
						["incomplete"] = L["Criteria Remaining (Zone)"],
						["totalComplete"] = L["Criteria Remaining (Total)"],
					},
				},
				combat = {
					type = "toggle",
					order = 2,
					name = L["Hide in Combat"],
					desc = L["Do not display the tooltip while in combat."],
				},
				dummy = {
					type = "description",
					order = 3,
					name = ""
				},
				location = {
					type = "toggle",
					order = 4,
					name = L["Location Info"],
					desc = L["Show location information in the LDB tooltip."],
				},
				remaining = {
					type = "toggle",
					order = 5,
					name = L["Criteria Remaining"],
					desc = L["Show the criteria remaining in the zone on the LDB tooltip."],
				},
				totalRemaining = {
					type = "toggle",
					order = 6,
					name = L["Total Remaining"],
					desc = L["Show the total remaining criteria for all zones on the LDB tooltip."],
				},
			}
		},
		frame = {
			type = "group",
			order = 100,
			name = L["Waypoint Frame"],
			get = function(info) return db.frame[info[#info]] end,
			set = function(info, value)
				db.frame[info[#info]] = value
				Lorewalkers:UpdateFrame()
				if waypoint then TomTom:RemoveWaypoint(waypoint) end
				if db.enabled then
					Lorewalkers:AddClosestWaypoint()
				end
			end,
			args = {
				enabled = {
					type = "toggle",
					order = 1,
					name = L["Enable Waypoint Frame"],
					desc = L["Toggle the waypoint frame."],
					get = function() return db.frame.enabled end,
					set = function(_, value)
						db.frame.enabled = value
						Lorewalkers:UpdateFrame()
						if not db.frame.enabled then
							LorewalkersFrame:Hide()
						end
						if waypoint then TomTom:RemoveWaypoint(waypoint) end
						if db.enabled then
							Lorewalkers:AddClosestWaypoint()
						end
					end,
				},
				frameSettings = {
					type = "group",
					order = 99,
					name = L["Frame Settings"],
					disabled = function() return not db.frame.enabled end,
					guiInline = true,
					args = {
						locked = {
							type = "toggle",
							order = 1,
							name = L["Locked"],
							get = function() return db.frame.locked end,
							set = function(_, value)
								db.frame.locked = value
								if db.frame.locked then
									LorewalkersFrame:Hide()
									LorewalkersFrame:SetScript("OnMouseDown", nil)
								else
									LorewalkersFrame:Show()
									LorewalkersFrame:SetScript("OnMouseDown", function() LorewalkersFrame:StartMoving() end)
								end
							end,
						},
						scale = {
							type = "range",
							order = 2,
							name = L["Scale"],
							min = 0.0, max = 3.0, step = 0.1,
						},
						alpha = {
							type = "range",
							order = 3,
							name = L["Transparency"],
							min = 0.0, max = 1.0, step = 0.1,
						},
					},
				},
			},
		},
		reset = {
			type = "group",
			order = 101,
			name = L["Manage Settings"],
			args = {
				waypoint = {
					type = "execute",
					order = 1,
					name = L["Reload Waypoint"],
					desc = L["Reload the current waypoint."],
					func = function()
						if waypoint then TomTom:RemoveWaypoint(waypoint) end
						Lorewalkers:AddClosestWaypoint()
					end,
				},
				criteria = {
					type = "execute",
					order = 2,
					name = L["Reload Criteria Info"],
					desc = L["Reloads criteria information into the database.  This will |cffff0000NOT|r reset the other settings."],
					func = function()
						Lorewalkers:LoadCriteriasIntoDatabase()
						Lorewalkers:Print(L["Reloaded criterias into the database."])
					end,
				},
				settings = {
					type = "execute",
					order = 3,
					name = L["Reset Settings"],
					desc = L["Reset the settings to their defaults."],
					func = function()
						db = defaults.char	-- set the database back to defaults
						Lorewalkers:LoadCriteriasIntoDatabase()
						if waypoint then TomTom:RemoveWaypoint(waypoint) end
						Lorewalkers:UpdateFrame()
						LorewalkersFrame:Show()
						Lorewalkers:AddClosestWaypoint()
					end,
				},
			},
		},
	},
}

-- ldb goodness
Lorewalkers.LDB = LibStub("LibDataBroker-1.1"):NewDataObject("Lorewalkers", {
	type = "data source",
	label = Lorewalkers.name,
	text = "0",
	icon = "Interface\\Icons\\achievement_faction_lorewalkers",
	OnClick = function(frame, button)
		if button == "LeftButton" then
			if not IsShiftKeyDown() then
				db.enabled = not db.enabled
				Lorewalkers:Print((L["Waypoints have been %s"]):format(db.enabled == true and enabledString or disabledString))
				if not db.enabled and waypoint then
					if LorewalkersFrame:IsShown() then LorewalkersFrame:Hide() end
					TomTom:RemoveWaypoint(waypoint)
				else
					if LorewalkersFrame:IsShown() then LorewalkersFrame:Hide() end
					Lorewalkers:UpdateCriteria()
					Lorewalkers:UpdateFrame()
					Lorewalkers:AddClosestWaypoint()
				end
			else
				InterfaceOptionsFrame_OpenToCategory(self.frames.general)
			end
		else
			db.frame.locked = not db.frame.locked
			Lorewalkers:Print((L["Waypoint Frame |cff00ff96%s|r"]):format(db.frame.locked == true and "Locked" or "Unlocked"))
			if db.frame.locked then
				LorewalkersFrame:Hide()
				LorewalkersFrame:SetScript("OnMouseDown", nil)
			else
				LorewalkersFrame:Show()
				LorewalkersFrame:SetScript("OnMouseDown", function() LorewalkersFrame:StartMoving() end)
			end
		end
	end,
})

local function InArray(needle, haystack)
	if #haystack == 0 then return false end
	for i = 1, #haystack do
		if haystack[i] == needle then
			return true
		end
	end
	return false
end

local function RepStatus()
	-- id 1345 are the lorewalkers
	local _, _, standing, barMin, barMax, barValue, _, _, _, _, _, _, _ = GetFactionInfoByID(1345)
	barValue = barValue - barMin
	barMax = barMax - barMin
	return barValue, barMax, repStandings[standing], math.floor(barValue / barMax * 100)
end

local function GetCriteriaCoordinates(criteria)
	for zone, waypoints in pairs(Lorewalkers.waypoints) do
		for name, info in pairs(waypoints) do
			if criteria == name then
				return zone, info.x, info.y
			end
		end
	end
	return false
end

function Lorewalkers:GetRemainingCriteria()
	local incomplete, complete, total = 0, 0, 0
	if db.ldb.display == "incomplete" or db.ldb.display == "complete" then
		local map = GetCurrentMapAreaID()
		if not db.waypoints[map] then return "N/A" end
		for _, point in pairs(db.waypoints[map]) do
			if point.completed == false then
				incomplete = incomplete + 1
			else
				complete = complete + 1
			end
			total = total + 1
		end
		return ("%d/%d"):format(db.ldb.display == "incomplete" and incomplete or complete, total)
	else
		for id, waypoints in pairs(db.waypoints) do
			for _, point in pairs(waypoints) do
				if point.completed == false then
					incomplete = incomplete + 1
				else
					complete = complete + 1
				end
				total = total + 1
			end
		end
		return ("%d/%d"):format(db.ldb.display == "totalIncomplete" and incomplete or complete, total)
	end
end

function Lorewalkers:LoadCriteriasIntoDatabase()
	db.waypoints = {}
	for mapID, waypoints in pairs(self.waypoints) do
		for cID, point in pairs(waypoints) do
			local _, achieveName, _, _, _, _, _, _, _, _, _, _, _ = GetAchievementInfo(point.achieveID)
			for i = 1, GetAchievementNumCriteria(point.achieveID) do
				local description, _, completed, _, _, _, _, _, _, criteriaID = GetAchievementCriteriaInfo(point.achieveID, i)
				if cID == criteriaID then
					if not db.waypoints[mapID] then db.waypoints[mapID] = {} end
					if not db.waypoints[mapID][cID] then db.waypoints[mapID][cID] = {} end
					db.waypoints[mapID][cID] = {
						achieveID = point.achieveID,
						achieveName = achieveName,
						name = description,
						note = point.note,
						num = i,
						x = point.x,
						y = point.y,
						completed = completed,
					}
				end
			end
		end
	end
end

function Lorewalkers:UpdateCriteria()
	if db.notices then
		self:Print(L["Updated achievement criterias."])
	end
	for mapID, waypoints in pairs(db.waypoints) do
		for cID, point in pairs(waypoints) do
			local _, _, completed, _, _, _, _, _, _, _ = GetAchievementCriteriaInfo(point.achieveID, point.num)
			db.waypoints[mapID][cID].completed = completed
		end
	end
end

function Lorewalkers:HandleChatCommand(args)
	local command, value = self:GetArgs(args, 2)
	local iconString = "\124T%s:12\124t"
	local completedString = "%s%s %s |cff00ff96%d/%d/%d|r"
	local notCompletedString = "%s%s |cffff0000%s|r"
	local criteriaCompleted = "    %d. %s: |cffffff00%s|r%s |cff00ff00%s|r"
	local criteriaNotCompleted = "    %d. %s: |cffffff00%s|r%s |cffff0000%s|r"
	local achievementPattern = "\124c%x+\124Hachievement:(%d-):.-\124h%[.-%]\124h\124r"
	
	if command == "toggle" then
		db.enabled = not db.enabled
		self:Print((L["Waypoints have been %s"]):format(db.enabled == true and enabledString or disabledString))
		if not db.enabled and waypoint then
			if LorewalkersFrame:IsShown() then LorewalkersFrame:Hide() end
			TomTom:RemoveWaypoint(waypoint)
		else
			if LorewalkersFrame:IsShown() then LorewalkersFrame:Hide() end
			self:UpdateCriteria()
			self:UpdateFrame()
			self:AddClosestWaypoint()
		end
		
	elseif command == "config" then
		InterfaceOptionsFrame_OpenToCategory(self.frames.general)
		
	elseif command == "rep" or command == "reputation" then
		local exaltedString = "%s |cff00ff96%d/%d %s (%d%%)|r %s |cff00ff96%s|r."
		local value, max, standing, percent = RepStatus()
		self:Print(exaltedString:format(L["You are"], value, max, standing, percent, L["with"], L["The Lorewalkers"]))
		
	elseif command == "criterias" or command == "criteria" or (value and value:match(achievementPattern)) then
		local id = 0
		if value:match(achievementPattern) then
			id = value:match(achievementPattern)
		elseif value:find(" ") then
			-- they provided a string, it must be in quotes
			if not Lorewalkers.achievements[value] then
				self:Print((L["Achievement |cffff0000\"%s\"|r doesn't exist."]):format(value))
				return
			else
				id = Lorewalkers.achievements[value]
			end
		else
			id = tonumber(value)
		end
		local link = GetAchievementLink(id)
		local _, _, _, completed, month, day, year, _, _, icon, _, _ = GetAchievementInfo(id)
		self:Print(completed == true and completedString:format(iconString:format(icon), link, L["Completed on"], month, day, year) or notCompletedString:format(iconString:format(icon), link, L["Incomplete"]))
		for i = 1, GetAchievementNumCriteria(id) do
			local description, _, completed, _, _, _, _, _, _, _ = GetAchievementCriteriaInfo(id, i)
			local zone, x, y = GetCriteriaCoordinates(description)
			if zone then
				print(completed == true and criteriaCompleted:format(i, description, zone, ("(%.2f, %.2f)"):format(x, y), L["Completed"]) or criteriaNotCompleted:format(i, description, zone, ("(%.2f, %.2f)"):format(x, y), L["Incomplete"]))
			else
				print(completed == true and criteriaCompleted:format(i, description, "", "", L["Completed"]) or criteriaNotCompleted:format(i, description, "", "", L["Incomplete"]))
			end
		end
	
	elseif command == "rlcriterias" or command == "rlcriteria" then
		self:LoadCriteriasIntoDatabase()
		if waypoint then TomTom:RemoveWaypoint(waypoint) end
		LorewalkersFrame:Hide()
		self:AddClosestWaypoint()
		self:Print(L["Reloaded criterias into the database."])
		
	elseif command == "achieve" or command == "achievements" then
		for _, id in pairs(self.achievements) do
			local link = GetAchievementLink(id)
			local _, _, _, completed, month, day, year, _, _, icon, _, _ = GetAchievementInfo(id)
			self:Print(completed == true and completedString:format(iconString:format(icon), link, L["Completed on"], month, day, year) or notCompletedString:format(iconString:format(icon), link, L["Incomplete"]))
		end
		
	elseif command == "clear" then
		if waypoint then TomTom:RemoveWaypoint(waypoint) end
	
	elseif command == "reload" then
		-- can't believe i havent added this until now
		if waypoint then TomTom:RemoveWaypoint(waypoint) end
		if LorewalkersFrame:IsShown() then LorewalkersFrame:Hide() end
		self:UpdateCriteria()
		self:UpdateFrame()
		self:AddClosestWaypoint()
		
	elseif command == "zone" then
		if GetCurrentMapContinent() ~= 6 then
			self:Print(L["You are not currently in Pandaria."])
		else
			local map, _, _, _ = Astrolabe:GetCurrentPlayerPosition()
			local complete, total = 0, 0
			if not db.waypoints[map] then
				self:Print(L["Something went wrong, please notify Lockslap."])
			else	
				for _, point in pairs(db.waypoints[map]) do
					if point.completed then
						complete = complete + 1
					end
					total = total + 1
				end
				self:Print((L["You have completed |cff00ff96%d|r of |cff00ff96%d (%d%%)|r in |cff00ff96%s|r."]):format(complete, total, math.ceil(complete / total * 100), GetMapNameByID(map)))
			end
		end
		
	elseif not command or command == "" or command == "help" or command == "?" then
		self:Print((L["Lorewalkers version |cff00ff96%.2f|r by |cff00ff00Lockslap|r"]):format(self.version))
		self:Print(L["Chat Command:  |cff00ff96/lorewalkers|r or |cff00ff96/lw|r"])
		print(L["    |cff00ff96config|r - Open Configuration Window"])
		print(L["    |cff00ff96clear|r - Clear Waypoints"])
		print(L["    |cff00ff96help|r or |cff00ff96?|r - Print This Help Again"])
		print(L["    |cff00ff96toggle|r - Toggle the Waypoints"])
		print(L["    |cff00ff96zone|r - Print the number of criteria remaining for your current zone."])
		print(L["    |cff00ff96rep|r or |cff00ff96reputation|r - Your Reputation with the Lorewalkers"])
		print(L["    |cff00ff96criteria|r or |cff00ff96criterias <id or name>|r - Display Achievement Criteria"])
		print(L["    |cff00ff96rlcriteria|r or |cff00ff96rlcriterias|r - Reload Criteria Information"]) 
		print(L["    |cff00ff96achievements|r or |cff00ff96achieve|r - Lorewalkers Achievement Progress"])
		
	end
end

function Lorewalkers:GetClosestWaypoint()
	local distance, point, criteriaID
	local map, level, x, y = Astrolabe:GetCurrentPlayerPosition()
	if not db.waypoints[map] or not map or not x or not y then return false end
	for criteria, points in pairs(db.waypoints[map]) do
		local wayDistance = Astrolabe:ComputeDistance(map, level, x, y, map, 0, points.x / 100, points.y / 100)
		if not points.completed then
			if not distance or wayDistance < distance then
				distance = wayDistance
				point = points
				criteriaID = criteria
			end
		end
	end
	-- we should have our nearest waypoint
	if not distance or not point then
		return false
	else
		return mapID, criteriaID, point, distance
	end
end

function Lorewalkers:AddClosestWaypoint()
	local zone = GetZoneText()
	local playerMap, _, _, _ = Astrolabe:GetCurrentPlayerPosition()
	if db.notices and GetCurrentMapContinent() == 6 and not db.waypoints[playerMap] then
		self:Print((L["No waypoints found for |cff00ff96%s|r."]):format(zone))
	elseif GetCurrentMapContinent() == 6 and db.enabled then
		local mapID, criteriaID, point, distance = self:GetClosestWaypoint()
		if not mapID or not criteriaID or not point or not distance then
			if db.notices then
				self:Print((L["You have completed all points for |cff00ff96%s|r"]):format(zone))
			end
		else
			if db.notices then
				self:Print((L["Loaded point for |cff00ff96%s|r, |cff00ff96%.2f|r yards away."]):format(point.name, distance))
			end
		
			-- thanks to Cladhaire, the author of TomTom, for pointing me in the right direction
			local opts = {
				title = ("%s\n%s"):format(point.name, point.achieveName),
				persistent = false,
				minimap = db.minimap,
				world = db.worldmap,
				crazy = db.arrow,
				cleardistance = 0,
				arrivaldistance = 0,
				callbacks = {},
			}
			opts.callbacks = TomTom:DefaultCallbacks(opts)
			opts.callbacks.distance[100] = function(event, uid, range, distance, lastDistance)
				if ((not lastDistance and distance <= 100) or (lastDistance > distance)) and db.frame.enabled then
					LorewalkersFrame:Show()
				elseif distance > 100 and LorewalkersFrame:IsShown() then
					LorewalkersFrame:Hide()
				end
			end
			waypoint = TomTom:AddMFWaypoint(mapID, 0, point.x / 100, point.y / 100, opts)
			
			if db.frame.enabled then
				self:UpdateFrame()
				self:SetScreenshot(mapID, criteriaID)
			end
		end
	end
end

function Lorewalkers:CRITERIA_EARNED(event, ...)
	-- loads the next waypoint when you complete a criteria
	local achievement, criteria = ...
	local zone, _, _, _ = Astrolabe:GetCurrentPlayerPosition()
	if achievement and criteria and db.waypoints[zone][criteria] then
		local _, _, completed, _, _, _, _, _, _, _, _ = GetAchievementCriteriaInfoByID(achievement, criteria)
		db.waypoints[zone][criteria].completed = completed
		
		if waypoint then
			TomTom:RemoveWaypoint(waypoint)
		end
		
		-- hide the screenshot frame, if necessary
		if LorewalkersFrame:IsShown() then
			LorewalkersFrame:Hide()
		end
		
		if GetCurrentMapContinent() == 6 and db.enabled then
			self:AddClosestWaypoint()
			-- print detailed info
			if db.info then
				self:UpdateCriteria()
				if db.infoType == "criteria" then
					local complete, total = 0, 0
					for i = 1, GetAchievementNumCriteria(achievement) do
						local _, _, completed, _, _, _, _, _, _, _ = GetAchievementCriteriaInfo(achievement, i)
						if completed then complete = complete + 1 end
						total = total + 1
					end
					self:Print((L["You have completed |cff00ff96%d|r of |cff00ff96%d (%d%%)|r for %s."]):format(complete, total, complete / total * 100, GetAchievementLink(achievement)))
				elseif db.infoType == "zone"then
					local complete, total = 0, #db.waypoints[zone]
					for _, point in pairs(db.waypoints[zone]) do
						if point.completed then complete = complete + 1 end
						total = total + 1
					end
					self:Print((L["You have completed |cff00ff96%d|r of |cff00ff96%d (%d%%)|r in |cff00ff96%s|r."]):format(complete, total, complete / total * 100, GetMapNameByID(zone)))
				end
			end
		end
	end
	
	Lorewalkers.LDB.text = self:GetRemainingCriteria() or 0
end
Lorewalkers:RegisterEvent("CRITERIA_EARNED")

function Lorewalkers:ZONE_CHANGED_NEW_AREA(event, ...)
	if waypoint then
		TomTom:RemoveWaypoint(waypoint)
		waypoint = nil
	end
	
	if GetCurrentMapContinent() == 6 and db.enabled then
		self:AddClosestWaypoint()
	elseif GetCurrentMapContinent() ~= 6 or not db.enabled then
		if LorewalkersFrame:IsShown() then
			LorewalkersFrame:Hide()
		end
	end
	
	Lorewalkers.LDB.text = self:GetRemainingCriteria() or 0
end
Lorewalkers:RegisterEvent("ZONE_CHANGED_NEW_AREA")

function Lorewalkers:PLAYER_ENTERING_WORLD()
	Lorewalkers.LDB.text = self:GetRemainingCriteria() or 0
	self:UnregisterEvent("PLAYER_ENTERING_WORLD")
end
Lorewalkers:RegisterEvent("PLAYER_ENTERING_WORLD")

function Lorewalkers:OnInitialize()
	-- load the config
	self.db = LibStub("AceDB-3.0"):New("LorewalkersDB", defaults)
	db = self.db.char
	
	-- setup tomtom
	TomTom.profile.arrow.enable = true
	TomTom.profile.arrow.arrival = 5
	TomTom.profile.arrow.setclosest = true
	
	-- this will be done the first time the addon is loaded
	if not db.firstLoad then
		self:LoadCriteriasIntoDatabase()
		db.firstLoad = true
	else
		self:UpdateCriteria()
	end
	
	-- add the about group
	options.args.about = {
		type = "group",
		order = 101,
		name = L["About"],
		args = {},
	}
	local i = 1
	for _, field in pairs(fields) do
		local val = GetAddOnMetadata(self.name, field)
		if val then
			options.args.about.args[field] = {
				type = "description",
				order = i,
				width = "normal",
				name = fieldName:format(field:gsub("X%-", "")),
			}
			i = i + 1
			options.args.about.args[("%s_value"):format(field)] = {
				type = "description",
				order = i,
				width = "double",
				name = val,
			}
			i = i + 1
		end
	end

	-- setup config window
	AceConfigRegistry:RegisterOptionsTable(self.name, options)
	self.frames = {}
	self.frames.general = AceConfigDialog:AddToBlizOptions(self.name, nil, nil, "general")
	self.frames.ldb = AceConfigDialog:AddToBlizOptions(self.name, L["LDB Plugin"], self.name, "ldb")
	self.frames.frame = AceConfigDialog:AddToBlizOptions(self.name, L["Waypoint Frame"], self.name, "frame")
	self.frames.reset = AceConfigDialog:AddToBlizOptions(self.name, L["Reset Settings"], self.name, "reset")
	self.frames.about = AceConfigDialog:AddToBlizOptions(self.name, L["About"], self.name, "about")
	
	-- setup the waypoint frame
	if db.frame.locked then
		LorewalkersFrame:Hide()
		LorewalkersFrame:SetScript("OnMouseDown", nil)
	else
		LorewalkersFrame:Show()
		LorewalkersFrame:SetScript("OnMouseDown", function() LorewalkersFrame:StartMoving() end)
	end
	self:UpdateFrame()
	LorewalkersFrameBG:SetTexture([[Interface\AddOns\Lorewalkers\Waypoints\Default.tga]])
end

function Lorewalkers:OnEnable()
	self:RegisterChatCommand("lorewalkers", "HandleChatCommand")
	self:RegisterChatCommand("lw", "HandleChatCommand")
	
	if db.enabled then
		self:ScheduleTimer("AddClosestWaypoint", 5)
	end
end

function Lorewalkers:OnDisable()
	self:UnregisterChatCommand("lorewalkers")
	self:UnregisterChatCommand("lw")
	
	if waypoint then
		TomTom:RemoveWaypoint(waypoint)
		waypoint = nil
	end
end

-------------------------------------------------------------------------------
-- Waypoint Screenshot Frame
-------------------------------------------------------------------------------
function Lorewalkers:SetScreenshot(zone, id)
	LorewalkersFrameBG:SetTexture(("Interface\\AddOns\\Lorewalkers\\Waypoints\\%d\\%d.tga"):format(zone, id))
end

function Lorewalkers:UpdateFrame()
	if not db.frame.enabled and LorewalkersFrame:IsShown() then
		LorewalkersFrame:Hide()
		return
	end
	LorewalkersFrame:SetHeight(128 * db.frame.scale)
	LorewalkersFrame:SetWidth(128 * db.frame.scale)
	LorewalkersFrame:SetAlpha(db.frame.alpha)
end

-------------------------------------------------------------------------------
-- LibDataBroker Stuff
-------------------------------------------------------------------------------
function Lorewalkers.LDB:OnEnter()
	local zone, subzone = GetZoneText(), GetSubZoneText()
	local map, level, x, y = Astrolabe:GetCurrentPlayerPosition()
	if db.ldb.combat and InCombatLockdown() then return end
	
	-- update the criteria in the database
	Lorewalkers:UpdateCriteria()

	-- setup the tooltip first
	GameTooltip:ClearLines()
	
	-- now build the content
	GameTooltip:AddLine((L["|cff00ff96%s|r v|cff00ff96%.2f|r by |cff9382c9Lockslap|r"]):format(Lorewalkers.name, Lorewalkers.version))
	GameTooltip:AddDoubleLine("    " .. L["Waypoints:"], db.enabled and enabledString or disabledString, 1, 1, 1, 1, 1, 1)
	
	-- location information
	if db.ldb.location then
		local subzone = 
		GameTooltip:AddDoubleLine("    " .. L["Current Zone:"], zone, 1, 1, 1, 0, 1, 0.59)
		if subzone and subzone:gsub(" ", "") ~= "" then
			GameTooltip:AddDoubleLine("    " .. L["Current Subzone:"], subzone, 1, 1, 1, 0, 1, 0.59)
		end
		GameTooltip:AddDoubleLine("    " .. L["Current Coordinates:"], ("%.2f, %.2f"):format(x * 100, y * 100), 1, 1, 1, 0, 1, 0.59)
	end
	
	-- criteria remaining for the zone
	if db.ldb.remaining then
		GameTooltip:AddLine(" ")
		if db.waypoints[map] then
			local disTable, todo = {}, false
			-- build remaining waypoints
			for _, point in pairs(db.waypoints[map]) do
				if point.completed == false then
					-- get the distance
					local distance = Astrolabe:ComputeDistance(map, level, x, y, map, 0, point.x / 100, point.y / 100)
					table.insert(disTable, {
						distance = distance,
						name = point.name,
					})
					todo = true
				end
			end
			
			if not todo then
				GameTooltip:AddLine((L["You have completed all the waypoints for |cff9382c9%s|r."]):format(zone))
			else
				GameTooltip:AddLine((L["Missing Criteria for |cff9382c9%s|r"]):format(zone))
				table.sort(disTable, function(a, b) return a.distance < b.distance end)
				for i = 1, #disTable do
					GameTooltip:AddDoubleLine("    " .. disTable[i].name, ("%.2fyds"):format(disTable[i].distance), 1, 1, 1, 0, 1, 0.59)
				end
			end
		else
			GameTooltip:AddLine((L["No waypoints for |cff9382c9%s|r!"]):format(zone))
		end
	end
	
	-- total criteria remaining
	if db.ldb.totalRemaining then
		GameTooltip:AddLine(" ")
		GameTooltip:AddLine(L["Total Waypoints Remaining"])
		local remaining, zonetotals = {}, {}
		local i, total, totalRemain = 1, 0, 0
		for id, waypoints in pairs(db.waypoints) do
			remaining[i] = {
				zone = id,
				remain = 0,
			}
			zonetotals[id] = 0
			for _, point in pairs(waypoints) do
				if point.completed == false then
					remaining[i].remain = remaining[i].remain + 1
					totalRemain = totalRemain + 1
				end
				total = total + 1
				zonetotals[id] = zonetotals[id] + 1
			end
			i = i + 1
		end
		table.sort(remaining, function(a, b) return a.remain > b.remain end)
		for i = 1, #remaining do
			--print(remaining[i].zone, GetMapNameByID(remaining[i].zone), remaining[i].remain, zonetotals[remaining[i].zone])
			GameTooltip:AddDoubleLine("    " .. GetMapNameByID(remaining[i].zone), ("%d/%d (%d%%)"):format(remaining[i].remain, zonetotals[remaining[i].zone], math.ceil(remaining[i].remain / zonetotals[remaining[i].zone] * 100)), 1, 1, 1, 0, 1, 0.59)
		end
		GameTooltip:AddDoubleLine("    " .. L["Total:"], ("%d/%d (%d%%)"):format(totalRemain, total, math.ceil(totalRemain / total * 100)), 1, 1, 0, 0, 1, 0.59)
	end
	
	-- add the hint
	GameTooltip:AddLine(" ")
	GameTooltip:AddDoubleLine(L["Left Click"], L["Toggle Waypoints"], 1, 1, 1, 1, 1, 0)
	GameTooltip:AddDoubleLine(L["Shift+Left Click"], L["Open Config"], 1, 1, 1, 1, 1, 0)
	GameTooltip:AddDoubleLine(L["Right Click"], L["Lock/Unlock Waypoint Frame"], 1, 1, 1, 1, 1, 0)
	
	-- finally show the tooltip
	GameTooltip:Show()
end

function Lorewalkers.LDB:OnLeave()
	GameTooltip:Hide()
end